
"""
GLOBAL LIST OF ALTERNATES AND SYNONYMS
"""

SHEETS = {
    "paleo": ["p", "paleo"],
    "chron": ["c", "chron"],
    "measurement": ["m", "meas", "measurement"],
    "ensemble": ["e", "ens", "ensemble"],
    "distribution": ["d", "dist", "distribution"],
    "summary": ["s", "sum", "summary"]
}

# TABLES
DATA_FRAMES = [
    "chron",
    "paleo",
    "ensemble",
    "calibratedage",
    "distribution",
    "model",
    "table",
    "measurement"
]

# Allowable int and float fields
NUMERIC_INTS = [
    "maxYear",
    "minYear",
    "pubYear",
    "number",
    "year",
    "age14c",
    "sd14c"
]

NUMERIC_FLOATS = [
    "LiPDVersion",
    "coordinates",
    "depth"
]

# FILTER
COMPARISONS = {
    "==": "=",
    "is": "=",
    "greater than": ">",
    "equals": "=",
    "equal": "=",
    "less than": "<"
}

# EXCEL
EXCEL_GEO = [
    "latMax",
    "latMin",
    "lonMax",
    "lonMin",
    "elevation",
    "siteName"
]


# Missing value name appears as many variations. Try to account for all of them
ALTS_MV = [
    'missing value',
    'missing values',
    'missingvalue',
    'missingvalues',
    'missing_values',
    'missing variables',
    'missing_variables',
    'missingvariables'
]

FUNDING_LIST = [
    'funding_agency_name',
    'grant'
]

# Use this list to sort the top level keys in the LiPD file.
# Ex: if the LiPD has geo data, take the whole geo dictionary and put it into section 8.
# reorganize(), create_blanks()
NOAA_ALL = {
    "Top": ['Study_Name', 'Online_Resource', "Online_Resource_Description", 'Original_Source_URL', 'Archive',
            "Dataset_DOI", "Parameter_Keywords"],
    "Contribution_Date": ['Date'],
    "File_Last_Modified_Date": ["Modified_Date"],
    "Title": ['Study_Name'],
    "Investigators": ['Investigators'],
    "Description_Notes_and_Keywords": ['Description'],
    "Publication": ['Author', 'Published_Date_or_Year', 'Published_Title', 'Journal_Name', 'Volume', 'Edition',
                    'Issue', 'Pages', "Report", 'DOI', 'Online_Resource', 'Full_Citation', 'Abstract'],
    "Funding_Agency": ["Funding_Agency_Name", "Grant"],
    "Site_Information": ["Site_Name", "Location", "Country", "Northernmost_Latitude", "Southernmost_Latitude",
                          "Easternmost_Longitude",  "Westernmost_Longitude", "Elevation"],
    "Data_Collection": ['Collection_Name', 'Earliest_Year', 'Most_Recent_Year', 'Time_Unit', 'Core_Length', 'Notes'],
    "Species": ['Species_Name', 'Species_Code', "Common_Name"],
    "Chronology_Information": ['Chronology'],
    "Variables": ["shortname", "what", "material", "error", "units", "seasonality", "archive", "detail", "method", "dataType"],
    "Data": ["Missing_Value"],
    # OTHERS. These are not sections, but they are important for parsing
    "Pub_Data_Citation": [""],
    # These are keys that do not currently have NOAA - LPD mappings
    "Ignore": ["Date", "Parameter_Keywords", "Dataset_Name", "Tree_Species_Code", "Species_Name", "Common_Name",
               "Core_Length", "Time_Unit", "Most_Recent_Year", "Earliest_Year", "Collection_Name",
               "Online_Resource_Description"]
}

UNITS = {
    "meters": "m",
    "meter": "m",
    "centimeters": "cm",
    "centimeter": "cm",
    "feet": "ft",
    "foot": "ft",
    "millimeter": "mm",
    "millimeters": "mm"
}

# LiPD on left, NOAA on right
# Used to map LiPD to NOAA ontology (and vice versa)
#  map_key()
NOAA_KEYS = {
    'datasetDOI': "Dataset_DOI",
    'LiPDVersion': 'LiPD_Version',
    'abstract': 'Abstract',
    'agency': 'Funding_Agency_Name',
    'archiveType': 'Archive',
    'authors': 'Authors',
    'citation': 'Full_Citation',
    'country': 'Country',
    'dataType': 'dataType',
    'description': 'what',
    'detail': 'detail',
    "funding": "Funding_Agency",
    "geo": "Site_Information",
    'grant': 'Grant',
    'identifier': 'DOI',
    'investigators': 'Investigators',
    'issue': 'Issue',
    'journal': 'Journal_Name',
    'location': 'Location',
    'measurementMaterial': 'material',
    'measurementMethod': 'method',
    'missingValue': 'Missing_Values',
    'notes': 'Description',
    'onlineResource': 'Online_Resource',
    "url": "Original_Source_URL",
    "originalDataURL": "Original_Source_URL",
    'originalSourceURL': 'Original_Source_URL',
    'pages': 'Pages',
    "pub": "Publication",
    'pubYear': 'Published_Date_or_Year',
    "year": "Published_Date_or_Year",
    'report': 'Report_Number',
    'seasonality': 'seasonality',
    'siteName': 'Site_Name',
    'studyName': 'Study_Name',
    'title': 'Published_Title',
    'uncertainty': 'error',
    'units': 'units',
    'variableName': 'shortname',
    'volume': 'Volume',
    ########################
    # MULTIPOINT
    # Since coordinates are not explicit individual fields in LiPD, we'll have to manually transfer these.
    # coordinates[0]: "Northernmost_Latitude",
    # coordinates[0]: "Southernmost_Latitude",
    # coordinates[1]: "Easternmost_Longitude",
    # coordinates[1]: "Westernmost_Longitude",
    # coordinates[2]: "Elevation", (OPTIONAL)
    ########################
    # SINGLE POINT
    # coordinates[0]: "Northernmost_Latitude",
    # coordinates[0]: "Southernmost_Latitude",
    # coordinates[1]: "Easternmost_Longitude",
    # coordinates[1]: "Westernmost_Longitude",
    # coordinates[2]: "Elevation", (OPTIONAL)
    ########################
    # "doi": pub[i]["identifier"][0]["id"]
}

# Excel on left, LiPD on right
# excel_main()
EXCEL_KEYS = {
    "archive type": "archiveType",
    "dataset name": "dataSetName",
    "metadata": "metadata",
    "chronology": "chronology",
    "data (qc)": "dataQC",
    "data(qc)": "dataQC",
    "data (original)": "dataOriginal",
    "data(original)": "dataOriginal",
    "data": "data",
    "proxyList": "proxy",
    "proxy": "proxy",
    "about": "about",
    "study title": "studyName",
    "investigators": "investigators",
    "authors": "author",
    "publication title": "title",
    "journal": "journal",
    "year": "year",
    "volume": "volume",
    "issue": "issue",
    "pages": "pages",
    "dataUrl": "dataUrl",
    "link": "link",
    "report number": "reportNumber",
    "doi": "id",
    "abstract": "abstract",
    "alternate citation": "citation",
    "site name": "siteName",
    "northernmost latitude": "latMax",
    "southernmost latitude": "latMin",
    "easternmost longitude": "lonMax",
    "westernmost longitude": "lonMin",
    "elevation": "elevation",
    "variablename": "variableName",
    "description": "description",
    "short_name": "variableName",
    "what": "description",
    "material": "measurementMaterial",
    "error": "error",
    "units": "units",
    "seasonality": "seasonality",
    "archive": "archive",
    "detail": "detail",
    "method": "method",
    "data_type": "dataType",
    "basis of climate relation": "basis",
    "climate_interpretation_code": "climateInterpretation",
    "climate_intepretation_code": "climateInterpretation",
    "notes": "notes",
    "comments": "notes"
}

EXCEL_TEMPLATE = [
    "data",
    "chronology table containing measured depths and ages",
    "chronology table as paragraph (insert table into cell b3) not preferred",
    "chronology table (use as many rows and columns are needed below this line)",
    "note: additional chronology tables should be put below the first one with no blank lines.",
    "variables",
    "use one row to define each variable, beginning with depth then age; add additional worksheets for additional tables",
    "note: data_type is 'n' for numeric and 'c' for character data",
    "paste data table below starting in column a",
    "the value or character string used as a placeholder for missing values",
    "<missing value>",
    "<notes that describe the table as a whole>"
]

# HEADER ROW for parsing excel worksheets
# excel_main()
EXCEL_HEADER = [
    "variablename",
    "short_name",
    "what",
    "material",
    "error",
    "units",
    "seasonality",
    "archive",
    "detail",
    "method",
    "data_type",
    "uncertainty_level",
    "calibration_curve",
    "climate_interpretation_code",
    "standard",
    "basis_of_climate_relation"
]




from collections import OrderedDict
import csv
import re
import os
import copy

from ..helpers.misc import remove_empty_doi, remove_empty_fields
from ..helpers.jsons import write_json_to_file
from ..helpers.bag import finish_bag
from ..helpers.alternates import NOAA_SITE_INFO, NOAA_IGNORE_KEYS, FUNDING_LIST, NOAA_ORDERING, UNITS, ALTS_MV
from ..helpers.regexes import re_var, re_var_split, re_name_unit, re_name_unit_range, re_doi, re_var_w_units, re_tab_split
from ..helpers.blanks import NOAA_EMPTY, NOAA_DATA_LINES, NOAA_VAR_LINES, EMPTY
from ..helpers.loggers import create_logger

logger_noaa_lpd = create_logger("noaa_lpd")


class NOAA_LPD(object):

    def __init__(self, dir_root, dir_tmp, name):
        self.dir_root = dir_root
        self.dir_tmp = dir_tmp
        self.dir_bag = os.path.join(dir_tmp, name)
        self.name = name
        self.name_txt = name + '.txt'
        self.name_lpd = name + '.lpd'
        self.metadata = {}

    def main(self):
        """
        Convert a NOAA text file into a lipds file.
        CSV files will be created if chronology or data sections are available.
        :return dict: Metadata Dictionary
        """
        logger_noaa_lpd.info("enter main")
        # Run the file through the parser
        # Sets self.metadata, Creates CSVs in dir_tmp
        os.chdir(self.dir_tmp)
        os.mkdir(self.name)
        os.chdir(self.dir_root)
        self.__parse()
        os.chdir(self.dir_bag)

        # Dump Metadata JSON to dir_tmp
        write_json_to_file(self.name + '.jsonld', self.metadata)

        # Create a bagit bag
        finish_bag(self.dir_bag)
        logger_noaa_lpd.info("exit main")
        return

    def __parse(self):
        """
        Parse
        Accept the text file. We'll open it, read it, and return a compiled dictionary to write to a json file
        May write a chronology CSV and a data CSV if those sections are available
        :return:
        """
        logger_noaa_lpd.info("enter parse")

        # Strings
        missing_str = ''
        data_filename = ''

        # Counters
        grant_id = 0
        funding_id = 0
        data_col_ct = 1
        line_num = 0

        # Boolean markers
        description_on = False
        publication_on = False
        abstract_on = False
        site_info_on = False
        chronology_on = False
        chron_vals_on = False
        variables_on = False
        data_vals_on = False
        data_on = False

        # Lists
        lat = []
        lon = []
        elev = []
        pub = []
        funding = []
        temp_abstract = []
        temp_description = []
        data_var_names = []
        data_col_list = []
        data_tables = []

        # All dictionaries needed to create JSON structure
        temp_funding = OrderedDict()
        temp_pub = OrderedDict()
        core_len = OrderedDict()
        geo_properties = OrderedDict()
        chron_dict = OrderedDict()
        data_dict_upper = OrderedDict()
        final_dict = OrderedDict()

        # Open the text file in read mode. We'll read one line at a time until EOF
        with open(self.name_txt, 'r') as f:
            logger_noaa_lpd.info("opened noaa file: {}".format(self.name_txt))
            for line in iter(f):
                line_num += 1

                # PUBLICATION
                # There can be multiple publications. Create a dictionary for each one.
                if publication_on:
                    # End of the section. Add the dictionary for this one publication to the overall list
                    if '-----' in line:
                        temp_pub = self.__reorganize_doi(temp_pub)
                        pub.append(temp_pub.copy())
                        temp_abstract.clear()
                        temp_pub.clear()
                        publication_on = False
                        logger_noaa_lpd.info("end section: Publication")
                    elif abstract_on:
                        # End of abstract: possibly more variables after.
                        if "#" in line:
                            abstract_on = False
                            temp_pub['abstract'] = ''.join(temp_abstract)
                            logger_noaa_lpd.info("end section: Abstract")
                            line = self.__str_cleanup(line)
                            key, value = self.__slice_key_val(line)
                            temp_pub[self.__camel_case(key)] = value
                        else:
                            temp_abstract.append(self.__str_cleanup(line))

                    # Add all info into the current publication dictionary
                    else:
                        line = self.__str_cleanup(line)
                        key, value = self.__slice_key_val(line)
                        if key in ("Author", "Authors"):
                            temp_pub["author"] = self.__reorganize_authors(value)
                        else:
                            temp_pub[self.__camel_case(key)] = value
                            if key == 'Abstract':
                                logger_noaa_lpd.info("reading section: Abstract")
                                abstract_on = True
                                temp_abstract.append(value)

                # DESCRIPTION AND NOTES
                # Descriptions are often long paragraphs spanning multiple lines, but don't follow the key/value format
                elif description_on:

                    # End of the section. Turn marker off and combine all the lines in the section
                    if '-------' in line:
                        description_on = False
                        value = ''.join(temp_description)
                        final_dict['description'] = value
                        logger_noaa_lpd.info("end section: Description_and_Notes")

                    # The first line in the section. Split into key, value
                    elif 'Description:' in line:
                        key, val = self.__slice_key_val(line)
                        temp_description.append(val)

                    # Keep a running list of all lines in the section
                    else:
                        line = self.__str_cleanup(line)
                        temp_description.append(line)

                # SITE INFORMATION (Geo)
                elif site_info_on:

                    if '-------' in line:
                        site_info_on = False
                        logger_noaa_lpd.info("end section: Site_Information")

                    else:
                        line = self.__str_cleanup(line)
                        key, value = self.__slice_key_val(line)
                        if key.lower() in NOAA_SITE_INFO['lat']:
                            lat.append(self.__convert_num(value))

                        elif key.lower() in NOAA_SITE_INFO['lon']:
                            lon.append(self.__convert_num(value))

                        elif key.lower() in NOAA_SITE_INFO['properties']:
                            if key.lower() == 'elevation':
                                val, unit = self.__split_name_unit(value)
                                elev.append(val)
                            else:
                                geo_properties[self.__camel_case(key)] = value

                # CHRONOLOGY
                elif chronology_on:
                    """
                    HOW IT WORKS:
                    Chronology will be started at "Chronology:" section header
                    Every line starting with a "#" will be ignored
                    The first line without a "#" is considered the variable header line. Variable names are parsed.
                    Each following line will be considered column data and sorted accordingly.
                    Once the "-----" barrier is reached, we exit the chronology section.
                    """
                    # When reaching the end of the chron section, set the marker to off and close the CSV file
                    if '-------' in line:
                        # Turn off markers to exit section
                        chronology_on = False
                        chron_vals_on = False
                        try:
                            # If nothing between the chronology start and the end barrier, then there won't be a CSV
                            if chron_start_line != line_num - 1:
                                try:
                                    chron_csv.close()
                                    logger_noaa_lpd.info("parse: chronology: no data found in chronology section")
                                except NameError:
                                    logger_noaa_lpd.debug(
                                        "parse: chronology_on: NameError: chron_csv ref before assignment, {}".format(
                                            self.name_txt))
                                    print(
                                        "Chronology section is incorrectly formatted. "
                                        "Section data will not be converted")

                            logger_noaa_lpd.info("end section: Chronology")
                        except NameError:
                            logger_noaa_lpd.debug(
                                "parse: chronology_on: NameError: chron_start_line ref before assignment, {}".format(
                                    self.name_txt))
                            print("Chronology section is incorrectly formatted. Section data will not be converted")

                    # Data values line. Split, then write to CSV file
                    elif chron_vals_on:
                        values = line.split()
                        try:
                            cw.writerow(values)
                        except NameError:
                            logger_noaa_lpd.debug(
                                "parse: chronology_on: NameError: csv writer ref before assignment, {}".format(
                                    self.name_txt))
                            print("Chronology section is incorrectly formatted. Section data will not be converted")

                    else:
                        try:
                            # Chron variable headers line
                            if line and line[0] != "#":
                                chron_filename = self.name + '.chron1.measurementTable1.csv'
                                # Organize the var header into a dictionary
                                variables = self.__reorganize_chron_header(line)

                                # Create a dictionary of info for each column
                                chron_col_list = self.__create_chron_cols(variables)
                                chron_dict['filename'] = chron_filename
                                chron_dict['chronTableName'] = 'Chronology'
                                chron_dict['columns'] = chron_col_list

                                # Open CSV for writing
                                csv_path = os.path.join(self.dir_bag, chron_filename)
                                chron_csv = open(csv_path, 'w+', newline='')
                                logger_noaa_lpd.info("opened csv file: {}".format(chron_filename))
                                cw = csv.writer(chron_csv)
                                # Turn the marker on to start processing the values columns
                                chron_vals_on = True
                        except IndexError:
                            logger_noaa_lpd.debug("parse: chronology: IndexError when attempting chron var header")

                # VARIABLES
                elif variables_on:
                    """
                    HOW IT WORKS:
                    Variable lines are the only lines that have a "##" in front of them.
                    Ignore all lines that don't match the "##" regex.
                    Once there's a match, start parsing the variable lines, and create a column entry for each line.
                    """
                    process_line = False
                    # End of the section. Turn marker off
                    if "------" in line:
                        variables_on = False
                        logger_noaa_lpd.info("end section: Variables")
                    for item in NOAA_VAR_LINES:
                        if item.lower() in line.lower():
                            process_line = False
                    for item in NOAA_EMPTY:
                        if item == line:
                            process_line = False
                    m = re.match(re_var, line)
                    if m:
                        process_line = True

                    # If the line isn't in the ignore list, then it's a variable line
                    if process_line:

                        # Split the line items, and cleanup
                        cleaned_line = self.__separate_data_vars(line)

                        # Add the items into a column dictionary
                        data_col_dict = self.__create_paleo_col(cleaned_line, data_col_ct)

                        # Keep a list of all variable names
                        try:
                            # Use this list later to cross check with the variable line in the Data section
                            data_var_names.append(data_col_dict['variableName'])
                        except KeyError:
                            data_var_names.append('')
                            logger_noaa_lpd.warn("parse: variables: "
                                                 "KeyError: {} not found in {}".format("variableName", "data_col_dict"))
                        # Add the column dictionary into a final dictionary
                        data_col_list.append(data_col_dict)
                        data_col_ct += 1

                # DATA
                # Missing Value, Create data columns, and output Data CSV
                elif data_on:
                    """
                    HOW IT WORKS:
                    Capture the "Missing Value" entry, if it exists.
                    Data lines should not have a "#" in front of them.
                    The first line without a "#" should be the variable header line
                    All lines that follow should have column data.
                    """
                    # Do not process blank or template lines
                    process_line = True
                    for item in NOAA_DATA_LINES:
                        if item in line:
                            process_line = False
                    for item in NOAA_EMPTY:
                        if item == line:
                            process_line = False
                    for item in ALTS_MV:
                        # Missing value found. Store entry
                        if item in line.lower():
                            process_line = False
                            line = self.__str_cleanup(line)
                            key, missing_str = self.__slice_key_val(line)

                    if process_line:
                        # Split the line at each space (There SHOULD one space between each variable. Not always true)
                        values = line.split()
                        # Write all data values to CSV
                        if data_vals_on:
                            try:
                                dw.writerow(values)
                            except NameError:
                                logger_noaa_lpd.debug(
                                    "parse: data_on: NameError: csv writer ref before assignment, {}".format(
                                        self.name_txt))

                        # Check for the line of variables
                        else:
                            var = self.__str_cleanup(values[0].lstrip())
                            # Check if a variable name is in the current line
                            if var.lower() in line.lower():
                                data_vals_on = True
                                logger_noaa_lpd.info("start section: Data_Values")
                                # Open CSV for writing
                                data_filename = self.name + 'paleoData1.measurementTable1.csv'
                                csv_path = os.path.join(self.dir_bag, data_filename)
                                data_csv = open(csv_path, 'w+', newline='')
                                logger_noaa_lpd.info("opened csv file: {}".format(data_filename))
                                dw = csv.writer(data_csv)

                # METADATA
                else:
                    # Line Continuation: Sometimes there are items that span a few lines.
                    # If this happens, we want to combine them all properly into one entry.
                    if '#' not in line and line not in NOAA_EMPTY and old_val:
                        if old_key in ('funding', 'agency'):
                            try:
                                temp_funding[old_key] = old_val + line
                            except KeyError as e:
                                logger_noaa_lpd.debug(
                                    "parse: metadata: line continuation: {} not found in {}, {}".format(old_key,
                                                                                                        "temp_funding",
                                                                                                        e))
                        else:
                            try:
                                final_dict[old_key] = old_val + line
                            except KeyError as e:
                                logger_noaa_lpd.debug(
                                    "parse: metadata: line continuation: {} not found in {}, {}".format(old_key,
                                                                                                        "temp_funding",
                                                                                                        e))

                    # No Line Continuation: This is the start or a new entry
                    else:
                        line = self.__str_cleanup(line)

                        # Grab the key and value from the current line
                        try:
                            # Split the line into key, value pieces
                            key, value = self.__slice_key_val(line)
                            l_key = key.lower()
                            cc_key= self.__camel_case(key)

                            # If there is no value, then we are at a section header.
                            # Data often has a blank value, so that is a special check.
                            if not value or l_key == 'data':
                                # Turn on markers if we run into section headers
                                if l_key == 'description_and_notes':
                                    description_on = True
                                    logger_noaa_lpd.info("reading section: Description_and_Notes")
                                elif l_key == 'publication':
                                    publication_on = True
                                    logger_noaa_lpd.info("reading section: Publication")
                                elif l_key == 'site_information':
                                    site_info_on = True
                                    logger_noaa_lpd.info("reading section: Site_Information")
                                elif l_key == 'chronology':
                                    chronology_on = True
                                    logger_noaa_lpd.info("reading section: Chronology")
                                    chron_start_line = line_num
                                elif l_key == 'variables':
                                    variables_on = True
                                    logger_noaa_lpd.info("reading section: Variables")
                                elif l_key == 'data':
                                    data_on = True
                                    logger_noaa_lpd.info("reading section: Data")
                            # For all
                            else:
                                # Ignore any entries that are specified in the skip list
                                if l_key not in NOAA_IGNORE_KEYS:
                                    # There can be multiple funding agencies and grants. Keep a list of dict entries
                                    if l_key in FUNDING_LIST:
                                        if l_key == 'funding_agency_name':
                                            funding_id += 1
                                            key = 'agency'
                                        elif l_key == 'grant':
                                            grant_id += 1
                                            key = 'grant'
                                        temp_funding[key] = value
                                        # If both counters are matching, we are ready to add content to the funding list
                                        if grant_id == funding_id:
                                            funding.append(temp_funding.copy())
                                            temp_funding.clear()
                                    else:
                                        # There's likely two "Online_Resource"s, and we need both, so check and concat
                                        if cc_key == "onlineResource":
                                            # If it exists, append. If not, add entry as a list
                                            if cc_key in final_dict:
                                                final_dict[cc_key].append(value)
                                            else:
                                                final_dict[cc_key] = [value]
                                        else:
                                            final_dict[cc_key] = value
                                    # Keep track of old key in case we have a line continuation
                                    old_key = key
                                    old_val = value.strip()
                        except TypeError as e:
                            logger_noaa_lpd.warn(
                                "parse: TypeError: none type received from slice_key_val, {}".format(e))

        # Wait to close the data CSV until we reached the end of the text file
        try:
            data_csv.close()
            logger_noaa_lpd.info("end section: Data_Values")
            logger_noaa_lpd.info("end section: Data")
        except NameError as e:
            print("Error: NOAA text file is contains format errors. Unable to process.")
            logger_noaa_lpd.debug(
                "parse: NameError: failed to close csv, invalid formatting in NOAA txt file, {}".format(e))

        # Piece together measurements block
        logger_noaa_lpd.info("compiling final paleoData")
        data_dict_upper['filename'] = data_filename
        data_dict_upper['paleoDataTableName'] = 'Data'
        data_dict_upper['missingValue'] = missing_str
        data_dict_upper['columns'] = data_col_list
        data_tables.append(data_dict_upper)

        # Piece together geo block
        logger_noaa_lpd.info("compiling final geo")
        geo = self.__create_coordinates(lat, lon, elev)
        geo['properties'] = geo_properties

        # Piece together final dictionary
        logger_noaa_lpd.info("compiling final master")
        final_dict['pub'] = pub
        final_dict['funding'] = funding
        final_dict['geo'] = geo
        final_dict['coreLength'] = core_len
        final_dict['chronData'] = [{"chronMeasurementTable": chron_dict}]
        final_dict['paleoData'] = data_tables
        self.metadata = final_dict
        logger_noaa_lpd.info("final dictionary compiled")

        # Start cleaning up the metadata
        logger_noaa_lpd.info("removing empty fields")
        self.metadata = remove_empty_fields(self.metadata)
        logger_noaa_lpd.info("removing empty doi")
        self.metadata = remove_empty_doi(self.metadata)
        logger_noaa_lpd.info("removing irrelevant keys")
        self.__remove_irr_fields()

        logger_noaa_lpd.info("exit parse")
        return

    @staticmethod
    def __create_paleo_col(l, col_count):
        """
        Receive split list from separate_data_vars, and turn it into a dictionary for that column
        :param list l:
        :param int col_count:
        :return dict:
        """
        # Format: what, material, error, units, seasonality, archive, detail, method,
        # C or N for Character or Numeric data, direction of relation to climate (positive or negative)
        d = OrderedDict()
        d['number'] = col_count
        for idx, var_name in enumerate(NOAA_ORDERING[11]):
            try:
                value = l[idx]
                # These two cases are nested in the column, so treat them special
                if var_name == "seasonality":
                    d["climateInterpretation"] = {var_name: value}
                elif var_name == "uncertainty":
                    d["calibration"] = {var_name: value}
                # All other cases are root items in the column, so add normally
                else:
                    d[var_name] = value
            except IndexError as e:
                logger_noaa_lpd.debug("create_var_col: IndexError: var: {}, {}".format(var_name, e))
        return d

    @staticmethod
    def __separate_data_vars(line):
        """
        For the variables section, clean up the line and return a list of each of the 10 items
        :param str line:
        :return str:
        """
        combine = []
        if '#' in line:
            line = line.replace("#", "")
            line = line.lstrip()
        if line not in NOAA_EMPTY and line not in EMPTY:
            m = re.match(re_var_split, line)
            if m:
                combine.append(m.group(1))
                attr = m.group(2).split(',')
                combine += attr
                for index, string in enumerate(combine):
                    combine[index] = string.strip()
                # for i, s in enumerate(combine):
                #     if not s or s in NOAA_EMPTY:
                #         del combine[i]
        return combine

    @staticmethod
    def __convert_num(number):
        """
        All path items are automatically strings. If you think it's an int or float, this attempts to convert it.
        :param str number:
        :return float or str:
        """
        try:
            return float(number)
        except ValueError as e:
            logger_noaa_lpd.warn("convert_num: ValueError: {}".format(e))
            return number

    @staticmethod
    def __camel_case(word):
        """
        Convert underscore naming into camel case naming
        :param str word:
        :return str:
        """
        word = word.lower()
        if '_' in word:
            split_word = word.split('_')
        else:
            split_word = word.split()
        if len(split_word) > 0:
            for i, word in enumerate(split_word):
                if i > 0:
                    split_word[i] = word.title()
        strings = ''.join(split_word)
        return strings

    @staticmethod
    def __name_unit_regex(word):
        """
        Split a name and unit that are bunched together (i.e. '250m')
        :param str word:
        :return str str:
        """
        value = ""
        unit = ""
        r = re.findall(re_name_unit, word)
        try:
            value = r[0][0]
        except IndexError as e:
            logger_noaa_lpd.warn("name_unit_regex: IndexError: value: {}, {}, {}".format(word, r, e))
        try:
            unit = r[0][1]
            # Replace unit with correct synonym.
            if unit.lower() in UNITS:
                unit = UNITS[unit]
        except IndexError as e:
            logger_noaa_lpd.warn("name_unit_regex: IndexError: unit: {}, {}, {}".format(word, r, e))
        if value:
            try:
                value = float(value)
            except ValueError as e:
                logger_noaa_lpd.warn("name_unit_regex: ValueError: val: {}, {}".format(value, e))
        return value, unit

    @staticmethod
    def __contains_digits(word):
        """
        Check if the string contains digits
        :param str word:
        :return:
        """
        return any(i.isdigit() for i in word)

    def __split_name_unit(self, line):
        """
        Split a string that has value and unit as one.
        :param str line:
        :return str str:
        """
        vals = []
        unit = ''
        if line != '' or line != ' ':
            # If there are parenthesis, remove them
            line = line.replace('(', '').replace(')', '')
            # When value and units are a range (i.e. '100 m - 200 m').
            if re.match(re_name_unit_range, line):
                m = re.findall(re_name_unit_range, line)
                if m:
                    for group in m:
                        for item in group:
                            try:
                                val = float(item)
                                vals.append(val)
                            except ValueError:
                                if item:
                                    unit = item
                # Piece the number range back together.
                if len(vals) == 1:
                    value = vals[0]
                else:
                    value = str(vals[0]) + ' to ' + str(vals[1])
            else:
                value, unit = self.__name_unit_regex(line)
            return value, unit

    @staticmethod
    def __str_cleanup(line):
        """
        Remove the unnecessary characters in the line that we don't want
        :param str line:
        :return str:
        """
        if '#' in line:
            line = line.replace("#", "")
            line = line.strip()
        if '-----------' in line:
            line = ''
        return line

    @staticmethod
    def __slice_key_val(line):
        """
        Get the key and value items from a line by looking for and lines that have a ":"
        :param str line:
        :return str str: Key, Value
        """
        position = line.find(":")
        # If value is -1, that means the item was not found in the string.
        if position != -1:
            key = line[:position]
            value = line[position + 1:]
            value = value.lstrip()
            return key, value
        else:
            key = line
            value = None
            return key, value

    def __create_coordinates(self, lat, lon, elev):
        """
        GeoJSON standard:
        Use to determine 2-point or 4-point coordinates
        :param list lat:
        :param list lon:
        :return dict:
        """
        # Sort lat an lon in numerical order
        lat.sort()
        lon.sort()
        geo_dict = {}
        # 4 coordinate values
        if len(lat) == 2 and len(lon) == 2:
            # Matching coordinate pairs. Not 4 unique values.
            if lat[0] == lat[1] and lon[0] == lon[1]:
                logger_noaa_lpd.info("coordinates found: {}".format("2"))
                lat.pop()
                lon.pop()
                geo_dict = self.__geo_point(lat, lon, elev)
            # 4 unique coordinates
            else:
                logger_noaa_lpd.info("coordinates found: {}".format("4"))
                geo_dict = self.__geo_multipoint(lat, lon, elev)
        # 2 coordinate values
        elif len(lat) == 1 and len(lon) == 1:
            logger_noaa_lpd.info("coordinates found: {}".format("2"))
            geo_dict = self.__geo_point(lat, lon, elev)
        # 0 coordinate values
        elif not lat and not lon:
            logger_noaa_lpd.info("coordinates found: {}".format("0"))
        else:
            geo_dict = {}
            logger_noaa_lpd.info("coordinates found: {}".format("too many"))
        return geo_dict

    @staticmethod
    def __geo_multipoint(lat, lon, elev):
        """
        GeoJSON standard:
        Create a geoJson MultiPoint-type dictionary
        :param list lat:
        :param list lon:
        :return dict:
        """
        logger_noaa_lpd.info("enter geo_multipoint")
        geo_dict = OrderedDict()
        geometry_dict = OrderedDict()
        coordinates = []
        # bbox = []
        temp = [None, None]
        # 4 unique values
        # # Creates bounding box
        # for index, point in enumerate(lat):
        #     bbox.append(lat[index])
        #     bbox.append(lon[index])

        # Creates coordinates list
        for i in lat:
            temp[0] = i
            for j in lon:
                temp[1] = j
                coordinates.append(copy.copy(temp))
        if elev:
            coordinates = coordinates + elev
        # Create geometry block
        geometry_dict['type'] = 'Polygon'
        geometry_dict['coordinates'] = coordinates
        # Create geo block
        geo_dict['type'] = 'Feature'
        # geo_dict['bbox'] = bbox
        geo_dict['geometry'] = geometry_dict

        return geo_dict

    @staticmethod
    def __geo_point(lat, lon, elev):
        """
        GeoJSON standard:
        Create a geoJson Point-type dictionary
        :param list lat:
        :param list lon:
        :return dict:
        """
        logger_noaa_lpd.info("enter geo_point")
        coordinates = []
        geo_dict = OrderedDict()
        geometry_dict = OrderedDict()
        for index, point in enumerate(lat):
            coordinates.append(lat[index])
            coordinates.append(lon[index])
        if elev:
            coordinates = coordinates + elev
        geometry_dict['type'] = 'Point'
        geometry_dict['coordinates'] = coordinates
        geo_dict['type'] = 'Feature'
        geo_dict['geometry'] = geometry_dict
        return geo_dict

    @staticmethod
    def __reorganize_doi(temp_pub):
        """
        Create a valid bib json entry for the DOI information.
        "DOI" is technically the only valid DOI key, but there are sometimes a doi_id and doi_url entry.
        Check for all three and compare them, then keep whichever seems best.
        :param dict temp_pub:
        :return dict:
        """
        doi_out = ""
        doi_list = []
        rm = ["doiId", "doi", "doiUrl"]
        # Check if both entries exist
        if "doi" and "doiId" in temp_pub:
            if temp_pub["doi"] == temp_pub["doiId"]:
                # If the entries are the same, then pick one to use
                doi_out = temp_pub["doiId"]
            else:
                # If entries are not the same, check if it matches the regex pattern.
                if re_doi.findall(temp_pub["doiId"]):
                    doi_out = temp_pub["doiId"]
                # If it doesnt match the regex, just use the "doi" entry as-is
                else:
                    doi_out = temp_pub["doi"]
        # Check if "doiId" entry exists. Most common entry.
        elif "doiId" in temp_pub:
            doi_out = temp_pub["doiId"]
        # Check if "doi" entry exists. Fallback.
        elif "doi" in temp_pub:
            doi_out = temp_pub["doi"]
        # It's unlikely that ONLY the doi_url exists, but check if no doi found so far
        elif "doiUrl" in temp_pub:
            doi_out = temp_pub["doiUrl"]

        # Log if the DOI is invalid or not.
        if not re_doi.match(doi_out):
            logger_noaa_lpd.info("reorganize_doi: invalid doi input from NOAA file")

        # Get a list of all DOIs found in the string
        matches = re.findall(re_doi, doi_out)
        logger_noaa_lpd.info("reorganize_dois: found {} dois: {}".format(len(matches), doi_out))

        # Add identifier block to publication dictionary
        for doi in matches:
            doi_list.append({"type": "doi", "id": doi})

        # Remove the other DOI entries
        for k in rm:
            try:
                del temp_pub[k]
            except KeyError:
                # If there's a KeyError, don't worry about it. It's likely that only one of these keys will be present.
                pass

        temp_pub["identifier"] = doi_list
        return temp_pub

    def __remove_irr_fields(self):
        """
        Remove NOAA keys that don't have a place in the LiPD schema
        :return:
        """
        for key in NOAA_IGNORE_KEYS:
            try:
                del self.metadata[key]
            except KeyError:
                # Key wasn't found. That's okay. No need for handling
                pass
        return



    @staticmethod
    def __create_chron_cols(metadata):
        """
        Use to collected chron metadata to create the chron columns
        :param dict metadata: key: variable, val: unit (optional)
        :return list: list of one dict per column
        """
        chron_col_list = []
        chron_col_ct = 1
        for variableName, unit in metadata.items():
            temp_dict = OrderedDict()
            temp_dict['column'] = chron_col_ct
            temp_dict['variableName'] = variableName
            temp_dict['unit'] = unit
            chron_col_list.append(copy.deepcopy(temp_dict))
            chron_col_ct += 1

        return chron_col_list

    @staticmethod
    def __reorganize_chron_header(line):
        """
        Reorganize the list of variables. If there are units given, log them.
        :param str line:
        :return dict: key: variable, val: units (optional)
        """
        d = {}
        # Header variables should be tab-delimited. Use regex to split by tabs
        m = re.split(re_tab_split, line)
        # If there was an output match from the line, then keep going
        if m:
            # Loop once for each variable in the line
            for s in m:
                # Match the variable to the 'variable (units)' regex to look for units
                m2 = re.match(re_var_w_units, s)
                # If there was a match
                if m2:
                    # If no units were found, set to blank
                    if m2.group(2) is None:
                        d[m2.group(1)] = ""
                    # Units were found
                    else:
                        # Set both values
                        d[m2.group(1)] = m2.group(2)
        return d

    @staticmethod
    def __reorganize_authors(authors):
        """
        Separate the string of authors and put it into a BibJSON compliant list
        :param str authors:
        :return list: List of dictionaries of author names.
        """
        # String SHOULD be semi-colon separated names.
        l = []
        s = authors.split(";")
        for author in s:
            try:
                l.append({"name": author.strip()})
            except AttributeError:
                logger_noaa_lpd.warning("reorganize_authors: AttributeError: authors incorrectly formatted")
        return l




"""
GLOBAL LIST OF EMPTIES AND IGNORED TERMS
"""

# GENERAL
EMPTY = ['', ' ', None, 'na', 'n/a', 'nan', '?', "'", "''", -999]

# NOAA
NOAA_EMPTY = ['\n', '', '#\n', '# \n', ' ']
NOAA_KEYS = ['earliest_year', 'most_recent_year', 'note']
NOAA_VAR_LINES = [
    'have no #',
    'variables',
    'lines begin with #',
    'double marker',
    'line variables format',
    'line format',
    'c or n for character or numeric',
    'preceded by'
]
NOAA_DATA_LINES = [
    'have no #',
    'tab-delimited text',
    'age ensembles archived'
]



import csv

from ..helpers.regexes import *
from ..helpers.directory import check_file_age
from ..helpers.google import get_google_csv
from .loggers import create_logger


logger_tslint = create_logger("ts_lint")

# VALIDATING AND UPDATING TSNAMES


def update_ts_names(metadata):
    """
    Updates the TS names for a given metadata dictionary
    :param dict metadata: Metadata dictionary
    :return dict: Updated metadata dictionary
    """
    full, quick = _fetch_ts_names()
    metadata = _verify_ts_names(full, quick, metadata)
    return metadata


def _fetch_ts_names():
    """
    Call down a current version of the TSNames spreadsheet from google. Convert to a structure better
    for comparisons.
    :return dict: Complete TSName dictionary. Keys: Valid TSName, Values: Synonyms
    :return list: List of valid TSnames
    """
    logger_tslint.info("enter fetch_tsnames")
    full = {"root": [], "pub": [], "climateInterpretation": [], "calibration": [], "geo": [], "paleoData": []}
    quick_list = []

    # Check if it's been longer than one day since updating the TSNames.csv file.
    # If so, go fetch the file from google in case it's been updated since.
    # Or if file isn't found at all, download it also.
    if check_file_age('tsnames.csv', 1):
        # Fetch TSNames sheet from google
        print("Fetching update for tsnames.csv")
        ts_id = '1C135kP-SRRGO331v9d8fqJfa3ydmkG2QQ5tiXEHj5us'
        get_google_csv(ts_id, 'tsnames.csv')
    try:
        logger_tslint.info("opening tsnames.csv")
        # Start sorting the tsnames into an organized structure
        with open('tsnames.csv', 'r') as f:
            r = csv.reader(f, delimiter=',')
            for idx, line in enumerate(r):
                if idx != 0:
                    # Do not record empty lines. Create list of non-empty entries.
                    line = [x for x in line if x]
                    # If line has content (i.e. not an empty line), then record it
                    if line:
                        # We don't need all the duplicates of pub and fund.
                        if "pub" in line[0] or "funding" in line[0]:
                            if re_pub_fetch.match(line[0]):
                                quick_list.append(line[0])
                                full['pub'].append(line)
                        elif re_misc_fetch.match(line[0]):
                            # Other Categories. Not special
                            quick_list.append(line[0])
                            cat, key = line[0].split('_')
                            full[cat].append(line)
                        else:
                            # Any of the root items
                            quick_list.append(line[0])
                            full["root"].append(line)
    except FileNotFoundError as e:
        logger_tslint.debug("fetch_tsnames: FileNotFound: tsnames.csv, {}".format(e))
        print("Failed to find TimeSeries lint file")
    logger_tslint.info("exit fetch_tsnames")
    return full, quick_list


def _verify_ts_names(full, quick_list, d):
    """
    Verify TSNames are current and valid. Compare to TSNames spreadsheet in Google Drive. Update where necessary.
    :param dict full: Complete TSName dictionary. Keys: Valid TSName, Values: Synonyms
    :param list quick_list: List of valid TSnames
    """
    logger_tslint.info("enter verify_tsnames")
    # Temp to store incorrect keys
    bad_keys = []
    limbo = {}

    # Build onto the "limbo" dictionary so we have a list of keys to replace.
    for k, v in d.items():
        # @context, pub, fund should be ignored
        if k not in quick_list and not re_pub_valid.match(k) and not re_fund_valid.match(k) and k != '@context':
            # Invalid key. Store in temp for processing.
            if k not in limbo:
                bad_keys.append(k)
    # Start to find replacements for empty entries in "limbo"
    for invalid in bad_keys:
        # Set incorrect name as key, and valid name as value.
        limbo[invalid] = _get_valid_ts_name(full, invalid)

    # Use limbo to start replacing entries in d
    for invalid, valid in limbo.items():
        try:
            # Add new item, and remove old item in one step
            d[valid] = d.pop(invalid)
        except KeyError as e:
            logger_tslint.warn("verify_tsnames: KeyError: valid: {}, invalid: {}, {}".format(invalid, valid, e))
    logger_tslint.info("exit verify_tsnames")
    return d


def _get_valid_ts_name(full, invalid):
    """
    Turn a bad tsname into a valid one.
    * Note: Index[0] for each TSName is the most current, valid entry. Index[1:] are synonyms
    :param str invalid: Invalid tsname
    :return str: valid tsname
    """
    valid = ''
    invalid = invalid.lower()
    try:
        # PUB ENTRIES
        if re_pub_invalid.match(invalid):

            # Case 1: pub1_year (number and hyphen)
            if re_pub_nh.match(invalid):
                s_invalid = invalid.split('_', 1)
                # Check which list the key word is in
                for line in full['pub']:
                    for key in line:
                        if s_invalid[1] in key.lower():
                            # Get the keyword from the valid entry.
                            v = line[0].split("_")
                            # Join our category with the valid keyword
                            valid = ''.join([s_invalid[0], '_', v[1]])

            # Case 2: pub_year (hyphen)
            elif re_pub_h.match(invalid):
                s_invalid = invalid.split('_', 1)
                # The missing pub number is the main problem, but the keyword may or may not be correct. Check.
                for line in full['pub']:
                    for key in line:
                        if s_invalid[1] in key.lower():
                            # We're going to use the valid entry as-is, because that's what we need for this case.
                            valid = line[0]

            # Case 3: pub1year (number)
            elif re_pub_n.match(invalid):
                s_invalid = re_pub_n.match(invalid)
                for line in full['pub']:
                    for key in line:
                        if s_invalid.group(2) in key.lower():
                            v = line[0].split('_', 1)
                            valid = ''.join(['pub', s_invalid.group(0), v[1]])

            # Case 4: pubYear (camelcase)
            elif re_pub_cc.match(invalid):
                valid = _iter_ts(full, 'pub', invalid)

        # FUNDING
        elif re_fund_invalid.match(invalid):
            if "grant" in invalid:
                valid = 'funding1_grant'
            elif "agency" in invalid:
                valid = "funding1_agency"

        # GEO
        elif re_geo_invalid.match(invalid):
            valid = _iter_ts(full, 'geo', invalid)

        # PALEODATA
        elif re_paleo_invalid.match(invalid):
            g1 = re_paleo_invalid.match(invalid).group(1)
            valid = _iter_ts(full, 'paleoData', g1)

        # CALIBRATION
        elif re_calib_invalid.match(invalid):
            g1 = re_calib_invalid.match(invalid).group(1)
            valid = _iter_ts(full, 'calibration', g1)

        # CLIMATE INTERPRETATION
        elif re_clim_invalid.match(invalid):
            g1 = re_clim_invalid.match(invalid).group(1)
            if 'climate' in g1:
                g1 = re.sub('climate', '', g1)
            valid = _iter_ts(full, 'climateInterpretation', g1)

        else:
            # ROOT
            valid = _iter_ts(full, 'root', invalid)

        # LAST CHANCE:
        # Specific case that isn't a typical format, or no match. Go through all possible entries.
        if not valid:
            valid = _iter_ts(full, None, invalid)

    except IndexError as e:
        logger_tslint.debug("get_valid_tsname: IndexError: getting ts name for {}, {}".format(invalid, e))

    if not valid:
        logger_tslint.debug("get_valid_tsname: valid ts name not found for: {}".format(invalid))
        return invalid

    return valid


def _iter_ts(full, category, invalid):
    """
    Match an invalid entry to one of the TSName synonyms.
    :param str category: Name of category being searched
    :param str invalid: Invalid tsname string
    :return str: Valid tsname
    """
    valid = ''

    # If a leading hyphen is in the string, get rid of it.
    if '_' == invalid[0]:
        invalid = invalid[1:]

    # If one specific category is passed through
    if category:
        for line in full[category]:
            for key in line:
                if invalid in key.lower():
                    valid = line[0]
                    break
    # If the entire TSNames dict is passed through (i.e. final effort, all categories have failed so far)
    else:
        for k, v in full.items():
            for line in v:
                for key in line:
                    if invalid in key.lower():
                        valid = line[0]
                        break

    return valid



from .regexes import re_filter_expr
from .alternates import COMPARISONS
from .loggers import create_logger
from .misc import match_operators, cast_float

logger_ts = create_logger("ts")


def translate_expression(expression):
    """
    Check if the expression is valid, then check turn it into an expression that can be used for filtering.
    :return list of lists: One or more matches. Each list has 3 strings.
    """
    logger_ts.info("enter translate_expression")
    m = re_filter_expr.findall(expression)
    matches = []
    if m:
        for i in m:
            logger_ts.info("parse match: {}".format(i))
            tmp = list(i[1:])
            if tmp[1] in COMPARISONS:
                tmp[1] = COMPARISONS[tmp[1]]
            tmp[0] = cast_float(tmp[0])
            tmp[2] = cast_float(tmp[2])
            matches.append(tmp)
    else:
        logger_ts.warn("translate_expression: invalid expression: {}".format(expression))
        print("Invalid input expression")
    logger_ts.info("exit translate_expression")
    return matches


def get_matches(expr_lst, ts):
    """
    Get a list of TimeSeries objects that match the given expression.
    :param list expr_lst: Expression
    :param dict ts: TimeSeries
    :return list names: Matched TimeSeries objects
    """
    logger_ts.info("enter get_matches")
    names = []
    match = False
    try:
        for ts_name, ts_data in ts.items():
            for expr in expr_lst:
                try:
                    val = ts_data[expr[0]]
                    # Check what comparison operator is being used
                    if expr[1] == 'in':
                        # "IN" operator can't be used in get_truth. Handle first.
                        if expr[2] in val:
                            match = True
                    elif match_operators(val, expr[1], expr[2]):
                        # If it's a typical operator, check with the truth test.
                        match = True
                    else:
                        # If one comparison is false, then it can't possibly be a match
                        match = False
                        break
                except KeyError as e:
                    logger_ts.warn("get_matches: KeyError: getting value from TimeSeries object, {}, {}".format(expr, e))
                    match = False
                except IndexError as e:
                    logger_ts.warn("get_matches: IndexError: getting value from TimeSeries object, {}, {}".format(expr, e))
                    match = False
            if match:
                names.append(ts_name)
    except AttributeError as e:
        logger_ts.debug("get_matches: AttributeError: invalid timeseries, {},  {}".format(type(ts), e))
        print("Error: Invalid TimeSeries")
    logger_ts.info("exit get_matches")
    return names


def extract_time_series(lipd_library, timeseries_library, convert):
    """
    Create a TimeSeries using the current files in LiPD_Library.
    :return obj: TimeSeries_Library
    """
    # Loop over the LiPD objects in the LiPD_Library
    for k, v in lipd_library.get_master().items():
        # Get metadata from this LiPD object. Convert it. Pass TSO metadata to the TS_Library.
        timeseries_library.loadTsos(v.get_name_ext(), convert.ts_extract_main(v.get_master()))


def export_time_series(lipd_library, timeseries_library, convert):
    """
    Export TimeSeries back to LiPD Library. Updates information in LiPD objects.
    """
    l = []
    # Get all TSOs from TS_Library, and add them to a list
    for k, v in timeseries_library.get_master().items():
        l.append({'name': v.get_lpd_name(), 'data': v.get_master()})
    # Send the TSOs list through to be converted. Then let the LiPD_Library load the metadata into itself.
    lipd_library.load_tsos(convert.lipd_extract_main(l))




import csv

from ..helpers.loggers import *
from ..helpers.misc import cast_values_csvs, cast_int

logger_csvs = create_logger("csvs")


# Merge CSV and Metadata


def merge_csv_metadata(d):
    """
    Using the given metadata dictionary, retrieve CSV data from CSV files, and insert the CSV
    values into their respective metadata columns. Checks for both paleoData and chronData tables.
    :param dict d: Metadata
    :return dict: Modified metadata dictionary
    """
    logger_csvs.info("enter merge_csv_metadata")

    # Add CSV to paleoData
    if "paleoData" in d:
        d["paleoData"] = _merge_csv_section(d["paleoData"], "paleo")

    # Add CSV to chronData
    if "chronData" in d:
        d["chronData"] = _merge_csv_section(d["chronData"], "chron")

    logger_csvs.info("exit merge_csv_metadata")
    return d


def _merge_csv_section(section_data, pc):
    """
    Add csv data to all paleo data tables
    :param dict section_data: Metadata
    :return dict: Modified metadata
    """
    logger_csvs.info("enter  merge_csv_section: {}".format(pc))

    try:
        # Loop through each table_data in paleoData
        for table_name, table_data in section_data.items():

            if "{}MeasurementTable".format(pc) in table_data:
                # Send whole table_data through. Adds csv data to columns
                for table_name2, table_data2 in table_data["{}MeasurementTable".format(pc)].items():
                    crumbs = "{}.{}.{}.{}".format(pc, table_name, "MeasurementTable", table_name2)
                    table_data["{}MeasurementTable".format(pc)][table_name2] = _add_csv_to_columns(table_data2, crumbs)

            if "{}Model".format(pc) in table_data:
                crumbs = "".format(pc, table_name, "Model")
                table_data["{}Model".format(pc)] = _merge_csv_model(table_data["{}Model".format(pc)], pc, crumbs)

    except AttributeError:
        print("Error: {} section must be a dictionary data type".format(pc))

    logger_csvs.info("exit  merge_csv_section: {}".format(pc))
    return section_data


def _merge_csv_model(models, pc, crumbs):
    """
    Add csv data to each column in chron model
    :param list models: Metadata
    :param str pc: Paleo or chron
    :param str crumbs: Hierarchy crumbs
    :return dict: Modified metadata
    """
    logger_csvs.info("enter merge_csv_model: {}".format(pc))

    try:
        for model in models:

            if "summaryTable" in model:
                crumbs2 = "{}.{}".format(crumbs, "summaryTable")
                model["summaryTable"] = _add_csv_to_columns(model["summaryTable"], crumbs2)

            if "{}ModelTable".format(pc) in model:
                crumbs2 = "{}.{}".format(crumbs, "ModelTable")
                model["{}ModelTable".format(pc)] = _add_csv_to_columns(model["{}ModelTable".format(pc)], crumbs2)

            if "ensembleTable" in model:
                crumbs2 = "{}.{}".format(crumbs, "ensembleTable")
                model["ensembleTable"] = _add_csv_to_columns(model["ensembleTable"], crumbs2)

            # Check for calibratedAges (old format) and distributionTable (current format)
            if "calibratedAges" in model:
                model["distributionTable"] = {}
                # Calibrated age tables are nested. Go down an extra layer.
                for k, v in model["calibratedAges"].items():
                    crumbs2 = "{}.{}.{}".format(crumbs, "distributionTable", k)
                    model["distributionTable"][k] = _add_csv_to_columns(v, crumbs2)
            elif "distributionTable" in model:
                for k, v in model["distributionTable"].items():
                    crumbs2 = "{}.{}.{}".format(crumbs, "distributionTable", k)
                    model["distributionTable"][k] = _add_csv_to_columns(v, crumbs2)
    except AttributeError:
        print("Error: Model section must be a list data type")

    logger_csvs.info("exit merge_csv_model: {}".format(pc))
    return models


def _add_csv_to_columns(table, crumbs):
    """
    Add csv data to each column in a list of columns
    :param dict table: Table metadata
    :param str crumbs: Hierarchy crumbs
    :return dict: Table metadata with csv "values" entry
    """
    # Get the filename or this table
    filename = _get_filename(table, crumbs)

    # If there's no filename, bypass whole process because there's no way to know which file to open
    if filename:
        # Call read_csv_to_columns for this filename. csv_data is list of lists.
        csv_data = _read_csv_from_file(filename)

        # Start putting CSV data into corresponding column "values" key
        try:
            for col_name, col in table['columns'].items():
                # The "number" entry in the ensemble table is a list of columns, instead of an int.
                # In this case, just store all the csv data as a batch.
                if isinstance(col["number"], list):
                    col["values"] = csv_data
                # For all other cases, "number" is a single int, and "values" should hold one column list.
                else:
                    col_num = cast_int(col["number"])
                    col['values'] = csv_data[col_num - 1]
        except IndexError:
            logger_csvs.warning("add_csv_to_columns: IndexError: index out of range of csv_data list")
        except KeyError:
            logger_csvs.debug("add_csv_to_columns: KeyError: missing columns key")
        except Exception as e:
            logger_csvs.debug("add_csv_to_columns: Unknown Error:  {}".format(e))
    return table


def _read_csv_from_file(filename):
    """
    Opens the target CSV file and creates a dictionary with one list for each CSV column.
    :param str filename:
    :return list of lists: column values
    """
    logger_csvs.info("enter read_csv_from_file")
    d = {}
    l = []
    try:
        with open(filename, 'r') as f:
            r = csv.reader(f, delimiter=',')

            # Create a dict with X lists corresponding to X columns
            for idx, col in enumerate(next(r)):
                d[idx] = []
                d = cast_values_csvs(d, idx, col)

            # Start iter through CSV data
            for row in r:
                for idx, col in enumerate(row):
                    # Append the cell to the correct column list
                    d = cast_values_csvs(d, idx, col)

        # Make a list of lists out of the dictionary instead
        for idx, col in d.items():
            l.append(col)
    except FileNotFoundError as e:
        print('CSV FileNotFound: ' + filename)
        logger_csvs.warn("read_csv_to_columns: FileNotFound: {}, {}".format(filename, e))
    logger_csvs.info("exit read_csv_from_file")
    return l


# SKIM CSV


def get_csv_from_metadata(name, metadata):
    """
    Two goals. Get all csv from metadata, and return new metadata with generated filenames to match files.
    :param str name: LiPD dataset name
    :param dict metadata: Metadata
    :return dict: Csv Data
    """
    logger_csvs.info("enter get_csv_from_metadata")

    _csv = {}
    _meta = metadata
    # crumbs = _get_dataset_name(metadata)

    if "paleoData" in metadata:
        # start building crumbs. This will be used to create new filenames.
        crumbs_tmp = "{}.Paleo".format(name)
        # Process paleoData section
        pd_meta, pd_csv = _get_csv_section(metadata["paleoData"], crumbs_tmp, "paleo")
        # Build up out master csv output data with our skimmed values
        _csv.update(pd_csv)
        # Set the new metadata in the output meta
        _meta["paleoData"] = pd_meta

    if "chronData" in metadata:
        crumbs_tmp = "{}.Chron".format(name)
        cd_meta, cd_csv = _get_csv_section(metadata["chronData"], crumbs_tmp, "chron")
        _csv.update(cd_csv)
        _meta["chronData"] = cd_meta

    logger_csvs.info("exit get_csv_from_metadata")
    return _meta, _csv


def _get_csv_section(_meta, crumbs, pc):
    """
    Get table name, variable name, and column values from paleo metadata
    :param dict _meta: Metadata
    :param str crumbs: Table name crumbs
    :param str pc: Paleo or chron
    :return dict: Metadata
    """
    logger_csvs.info("enter get_csv_section: {}".format(pc))
    _csv = {}
    s_idx = 1
    try:
        # Process the tables in section
        for name_table, data_table in _meta.items():
            # crumbs_tmp = "{}{}".format(crumbs, str(name_table))
            crumbs_tmp = "{}{}".format(crumbs, s_idx)

            # Process each entry sub-table below if they exist
            if "{}MeasurementTable".format(pc) in data_table:
                idx = 1
                for name_pmt, dat in data_table["{}MeasurementTable".format(pc)].items():
                    # String together the final pieces of the crumbs filename
                    filename = "{}.{}{}.csv".format(crumbs_tmp, "measurementTable", idx)
                    # Set the filename inside the metadata also, so our _csv and _meta will match
                    dat = _set_filename(dat, filename)
                    # Get a nested list of table values
                    out = _search_table_for_vals(dat, filename)
                    # Set the table values to _csv using our filename.
                    _csv[filename] = out
                    idx += 1

            if "{}Model".format(pc) in data_table:
                for item in data_table["{}Model".format(pc)]:
                    if "calibratedAges" in item:
                        idx = 1
                        # CA has an extra level of nesting
                        for name_ca, dat in item["calibratedAges"].items():
                            filename = "{}.distributionTable{}.csv".format(crumbs_tmp, idx)
                            dat = _set_filename(dat, filename)
                            out = _search_table_for_vals(dat, filename)
                            _csv[filename] = out
                            idx += 1

                    elif "distributionTable" in item:
                        idx = 1
                        # CA has an extra level of nesting
                        for name_ca, dat in item["distributionTable"].items():
                            filename = "{}.distributionTable{}.csv".format(crumbs_tmp, idx)
                            dat = _set_filename(dat, filename)
                            out = _search_table_for_vals(dat, filename)
                            _csv[filename] = out
                            idx += 1

                    if "{}ModelTable".format(pc) in item:
                        filename = "{}.{}.csv".format(crumbs_tmp, "{}ModelTable".format(pc))
                        item["{}ModelTable".format(pc)] = _set_filename(item["{}ModelTable".format(pc)], filename)
                        out = _search_table_for_vals(item["{}ModelTable".format(pc)], filename)
                        _csv[filename] = out

                    elif "summaryTable".format() in item:
                        filename = "{}.{}.csv".format(crumbs_tmp, "summaryTable".format(pc))
                        item["summaryTable"] = _set_filename(item["summaryTable"], filename)
                        out = _search_table_for_vals(item["summaryTable"], filename)
                        _csv[filename] = out

                    if "ensembleTable" in item:
                        filename = "{}.{}.csv".format(crumbs_tmp, "ensembleTable")
                        item["ensembleTable"] = _set_filename(item["ensembleTable"], filename)
                        out = _search_table_for_vals(item["ensembleTable"], filename)
                        _csv[filename] = out
            s_idx += 1

    except AttributeError:
        logger_csvs.info("get_csv_section: {}, AttributeError: expected type dict, given type {}".format(pc, type(_meta)))

    logger_csvs.info("exit get_csv_section: {}".format(pc))
    return _meta, _csv


def write_csv_to_file(d):
    """
    Writes columns of data to a target CSV file.
    :param dict d: A dictionary containing one list for every data column. Keys: int, Values: list
    :return None:
    """
    logger_csvs.info("enter write_csv_to_file")
    for filename, data in d.items():
        l_columns = []
        for k, v in data.items():
            l_columns.append(v)
        rows = zip(*l_columns)
        with open(filename, 'w+') as f:
            w = csv.writer(f)
            for row in rows:
                w.writerow(row)
    logger_csvs.info("exit write_csv_to_file")
    return


# HELPERS


def _search_table_for_vals(d, filename):
    """
    Search a data tables for column values. Return a dict of column values
    :param dict d: Table data
    :return dict: Column values. ref by var name
    """
    cols = {}
    # if something goes wrong, we want a filename that we can track this data back to
    # however, if there's no "filename" entry, then we'll leave that data missing
    if "columns" in d:
        try:
            for name_col, data_col in d["columns"].items():
                vals = _search_col_for_vals(data_col, filename)
                if vals:
                    cols[name_col] = vals
        except AttributeError:
            print("Error: Table 'columns' entries must be a dictionary type")
            logger_csvs.debug("search_table_for_vals: AttributeError: expected type dict, given type {}".format(type(d)))

    return cols


def _search_col_for_vals(data_col, filename):
    """
    Get the values key from a data column
    :param dict data_col: Metadata column w/ values
    :param str filename: Filename of where this column came from
    :return list:
    """
    val = []
    try:
        val = data_col["values"]
    except KeyError:
        # tell user which csv file and variable is causing issues.
        print("Error: {}, '{}' is missing a 'values' entry".format(filename, data_col["variableName"]))
        logger_csvs.info("_search_col_for_vals: KeyError: Filename: {},  '{}' is missing a 'values' "
                         "entry".format(filename, data_col["variableName"]))

    return val


def _set_filename(table, filename):
    """
    Overwrite filename in table with our standardized filename
    :param dict table: Metadata
    :param str filename: Crumbs filename
    :return dict: Metadata
    """
    try:
        table["filename"] = filename
    except KeyError:
        logger_csvs.info("_set_filename: KeyError, Unable to set filename into table")
    return table


def _get_filename(table, crumbs):
    """
    Get the filename from a data table. If it doesn't exist, create a new one based on table hierarchy in metadata file.
    format: <dataSetName>.<dataTableType><idx>.<dataTableName>.csv
    example: ODP1098B.Chron1.ChronMeasurementTable.csv
    :param dict table: Table data
    :return str: Filename
    """
    try:
        filename = table["filename"]
    except KeyError:
        logger_csvs.info("_get_filename: KeyError: missing filename key for {}".format(crumbs))
        print("Error: Missing filename for: {} , cannot load this file".format(crumbs))
        filename = ""
    return filename


def _get_dataset_name(d):
    """
    Get data set name from metadata
    :param dict d: Metadata
    :return str: Data set name
    """
    try:
        s = d["dataSetName"]
    except KeyError:
        logger_csvs.warn("get_dataset_name: KeyError: missing dataSetName")
        s = "lipds"
    return s


def _merge_ensemble(ensemble, col_nums, col_vals):
    """
    The second column is not typical.
    "number" is a list of column numbers and "values" is an array of column values.
    Before we can write this to csv, it has to match the format the writer expects.
    :param dict ensemble: First column data
    :param list col_nums: Second column numbers. list of ints
    :param list col_vals: Second column values. list of lists
    :return dict:
    """
    try:
        # Loop for each column available
        for num in col_nums:
            # first column number in col_nums is usually "2", so backtrack once since ensemble already has one entry
            # col_vals is 0-indexed, so backtrack 2 entries
            ensemble[num-1] = col_vals[num - 2]

    except IndexError:
        logger_csvs.debug("merge_ensemble: IndexError: index out of range")

    return ensemble




import os
import shutil

from ..helpers.misc import update_lipd_version
from ..helpers.directory import create_tmp_dir
from ..helpers.jsons import read_json_from_file
from ..helpers.zips import zipper, unzipper
from .lpd_noaa import LPD_NOAA
# from .noaa_lpd import NOAA_LPD
from ..helpers.loggers import create_logger

logger_noaa = create_logger("noaa")


def noaa_main(files):
    """
    Convert between NOAA and LiPD file formats.
    :return:
    """
    logger_noaa.info("enter noaa")
    # Run lpd_noaa or noaa_lpd ?
    print("Which conversion?\n1. LPD to NOAA\n2. NOAA to LPD\n")
    mode = input("Option: ")
    logger_noaa.info("chose option: {}".format(mode))
    ft = ""
    ft_print = ""

    # .lpd to noaa
    if mode == '1':
        ft = ".lpd"
        ft_print = ' LiPD'
    # Find all needed files in current directory
    elif mode == '2':
        ft = ".txt"
        ft_print = ' NOAA'
    # only continue if the user selected a mode correctly
    if ft and ft_print:
        logger_noaa.info("Found {} {} file(s)".format(str(len(files[ft])), ft_print))
        print("Found {0} {1} file(s)".format(str(len(files[ft])), ft_print))
        # Process each available file of the specified .lpd or .txt type
        for file in files[ft]:
            # try to filter out example files and stuff without real data
            if "template" not in file["filename_ext"] and "example" not in file["filename_ext"]:
                os.chdir(file["dir"])
                print('processing: {}'.format(file["filename_ext"]))
                logger_noaa.info("processing: {}".format(file["filename_ext"]))

                # Unzip file and get tmp directory path
                dir_tmp = create_tmp_dir()

                # Process file
                if mode == '1':
                    unzipper(file["filename_ext"], dir_tmp)
                    _process_lpd(file, dir_tmp)
                elif mode == '2':
                    _process_noaa(file, dir_tmp)

                # Delete tmp folder and all contents
                os.chdir(file["dir"])
                shutil.rmtree(dir_tmp)
    else:
        print("Error: a mode was not selected")
        logger_noaa.info("noaa_main: a mode was not selected")
    print("Process Complete")
    logger_noaa.info("exit noaa")
    return


def _process_noaa(file, dir_tmp):
    """
    Convert NOAA format to LiPD format
    :param dict file: File metadata
    :param str dir_tmp: Path to tmp directory
    :return None:
    """
    logger_noaa.info("enter process_noaa")
    # NOAA_LPD(dir_root, dir_tmp, name).main()
    os.chdir(file["dir"])
    zipper(dir_tmp, file["filename_no_ext"], file["filename_ext"])
    logger_noaa.info("exit process_noaa")
    return


def _process_lpd(file, dir_tmp):
    """
    Convert a LiPD format to NOAA format
    :param dict file: File metadata
    :param str dir_tmp: Path to tmp directory
    :return None:
    """
    logger_noaa.info("enter process_lpd")
    dir_bag = os.path.join(dir_tmp, file["filename_no_ext"])
    dir_data = os.path.join(dir_bag, 'data')

    # Navigate down to jld file
    # Directory Change: dir_root -> dir_data
    try:
        os.chdir(dir_data)
        # Open file and execute conversion script
        d = read_json_from_file(os.path.join(dir_data, file["filename_no_ext"] + '.jsonld'))
        # Do we need to update json to most recent LiPD Version? Check before passing to converter
        d = update_lipd_version(d)
        # create object and start conversion process
        LPD_NOAA(file["dir"], file["filename_no_ext"], d).main()
    except FileNotFoundError:
        logger_noaa.debug("process_lpd: FileNotFound: tmp directory not found")
        print("Error: Unable to process {}".format(file["filename_no_ext"]))
    logger_noaa.info("exit process_lpd")
    return




import zipfile
import shutil
import os

from .loggers import create_logger

logger_zips = create_logger("zips")


def zipper(dir_tmp, name, name_ext):
    """
    Zips up directory back to the original location
    :param str dir_tmp: Path to tmp directory
    :param str name: Name of current file
    :param str name_ext: Name of current file with '.lpd' extension
    """
    logger_zips.info("re_zip: name: {}, dir_tmp: {}".format(name_ext, dir_tmp))
    # creates a zip archive in current directory. "somefile.lpd.zip"
    shutil.make_archive(name_ext, format='zip', root_dir=dir_tmp, base_dir=name)
    # drop the .zip extension. only keep .lpd
    os.rename("{}.zip".format(name_ext), name_ext)
    return


def unzipper(name_ext, dir_tmp):
    """
    Unzip .lpd file contents to tmp directory.
    :param str name_ext: Name of lpd file with extension
    :param str dir_tmp: Tmp folder to extract contents to
    :return None:
    """
    logger_zips.info("enter unzip")
    # Unzip contents to the tmp directory
    try:
        with zipfile.ZipFile(name_ext) as f:
            f.extractall(dir_tmp)
    except FileNotFoundError as e:
        logger_zips.debug("unzip: FileNotFound: {}, {}".format(name_ext, e))
        shutil.rmtree(dir_tmp)
    logger_zips.info("exit unzip")
    return



from .doi_resolver import DOIResolver
from ..helpers.bag import *
from ..helpers.directory import *
from ..helpers.zips import *
from ..helpers.jsons import *
from ..helpers.loggers import create_logger

logger_doi_main = create_logger("doi_main")


def doi_main(files):
    """
    Main function that controls the script. Take in directory containing the .lpd file(s). Loop for each file.
    :return None:
    """
    logger_doi_main.info("enter doi_main")

    print("Found {0} {1} file(s)".format(str(len(files["lipd"])), 'LiPD'))
    force = prompt_force()
    for file in files["lipd"]:

        # Unzip file and get tmp directory path
        dir_tmp = create_tmp_dir()
        unzipper(file["filename_ext"], dir_tmp)

        # Force DOI update?
        if force:
            # Update file. Forcing updates for all files.
            print('processing: {}'.format(file["filename_ext"]))
            logger_doi_main.info("processing: {}".format(file["filename_ext"]))
            # dir: dir_root -> dir_tmp
            process_lpd(file["filename_no_ext"], dir_tmp)
            # dir: dir_tmp -> dir_root
            os.chdir(file["dir"])
            # Zip the directory containing the updated files. Created in dir_root directory
            zipper(dir_tmp, file["filename_no_ext"], file["filename_ext"])
            # Cleanup and remove tmp directory
            shutil.rmtree(dir_tmp)

        if not force:
            # Don't Update File. Flag found and we're not forcing updates.
            if resolved_flag(open_bag(os.path.join(dir_tmp, file["filename_no_ext"]))):
                print('skipping: {}'.format(file["filename_ext"]))
                logger_doi_main.info("skipping: {}".format(file["filename_ext"]))
                shutil.rmtree(dir_tmp)

            # Update File. No flag found and hasn't been processed before.
            else:
                print('processing: {}'.format(file["filename_ext"]))
                logger_doi_main.info("processing: {}".format(file["filename_ext"]))
                # dir: dir_root -> dir_tmp
                process_lpd(file["filename_no_ext"], dir_tmp)
                # dir: dir_tmp -> dir_root
                os.chdir(file["dir"])
                # Zip the directory containing the updated files. Created in dir_root directory
                zipper(dir_tmp, file["filename_no_ext"], file["filename_ext"])
                # Cleanup and remove tmp directory
                shutil.rmtree(dir_tmp)
    logger_doi_main.info("exit doi_main")
    print("Process Complete")
    return


def process_lpd(name, dir_tmp):
    """
    Opens up json file, invokes doi_resolver, closes file, updates changelog, cleans directory, and makes new bag.
    :param str name: Name of current .lpd file
    :param str dir_tmp: Path to tmp directory
    :return none:
    """
    logger_doi_main.info("enter process_lpd")
    dir_root = os.getcwd()
    dir_bag = os.path.join(dir_tmp, name)
    dir_data = os.path.join(dir_bag, 'data')

    # Navigate down to jLD file
    # dir : dir_root -> dir_data
    os.chdir(dir_data)

    # Open jld file and read in the contents. Execute DOI Resolver.
    jld_data = read_json_from_file(os.path.join(dir_data, name + '.jsonld'))

    # Overwrite data with new data
    jld_data = DOIResolver(dir_root, name, jld_data).main()
    # Open the jld file and overwrite the contents with the new data.
    write_json_to_file(os.path.join(dir_data, name + '.jsonld'), jld_data)

    # Open changelog. timestamp it. Prompt user for short description of changes. Close and save
    update_changelog()

    # Delete old bag files, and move files to bag root for re-bagging
    # dir : dir_data -> dir_bag
    dir_cleanup(dir_bag, dir_data)
    finish_bag(dir_bag)

    logger_doi_main.info("exit process_lpd")
    return


def prompt_force():
    """
    Ask the user if they want to force update files that were previously resolved
    :return bool: response
    """
    logger_doi_main.info("enter prompt_force")
    count = 0
    print("Do you want to force updates for previously resolved files? (y/n)")
    while True:
        force = input("> ")
        try:
            if count == 2:
                return True
            elif force.lower() in ('y', 'yes'):
                return True
            elif force.lower() in ('n', 'no'):
                return False
            else:
                print("invalid response")
        except AttributeError as e:
            print("invalid response")
            logger_doi_main.warn("invalid response: {}, {}".format(force, e))
        count += 1
    logger_doi_main.info("force update: {}".format(force))
    logger_doi_main.info("exit prompt_force")
    return True

if __name__ == '__main__':
    doi_main()
